/*
The GPLv3 License (GPLv3)

Copyright (c) 2022 Akos Horvath

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef WM_H
#define WM_H

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xinerama.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "config.h"
#include "util.h"

// TODO: when only 1 window on ws it is not focused
// TODO: dont draw border on not normal windows, floating dialog window,
//       window type/name/... atom member variable on Client struct
// TODO: get_atom probably not working correctly
// TODO: important: code cleanliness, ewmh!!
// TODO: config command line utility(sockets), maybe avoid global vars

#define RETURN_IF_NULL(c)   \
if (c == NULL)              \
{                           \
    fprintf(stderr, "wm: variable %s was null in function %s\n", #c, __func__); \
    return;                 \
}

#ifdef DEBUG
#define DEBUG_PRINT(fmt, ...) \
    do { { fprintf(stderr, "[%s:%d] ", __FILE__, __LINE__); \
           fprintf(stderr, fmt, ##__VA_ARGS__);} } while (0);
#else
#define DEBUG_PRINT(fmt, ...)
#endif

#define WM_WS_NAME_LEN 64

typedef struct Client Client;
typedef struct Workspace Workspace;
typedef struct Monitor Monitor;
typedef struct Arg Arg;
typedef struct Wm Wm;

enum Direction {
    UP,
    DOWN,
    LEFT,
    RIGHT
};

typedef enum {
    SPLIT_VERTICAL,
    SPLIT_HORIZ,
    SPLIT_TAB
} SplitMode;

struct Monitor {
    Display *display;
    XineramaScreenInfo info;
    Client *clients;
    Workspace *workspaces;
    size_t selws;
    size_t wscount;
};

struct Workspace {
    TreeNode *tree;
    size_t index;
    char name[WM_WS_NAME_LEN];
    Monitor *monitor;
    SplitMode split;
};

struct Wm {
    // TODO INIT
    Display *display;
    Monitor *monitors;
    Monitor *smon;
    bool running;
    int wm_mc;
    bool other_wm;
    Client root;
    Window dock;

    Client *focused_client;

    // float cfg_ms_p;
    // char *cfg_border_col;
    // char *cfg_focused_border_col;
    // unsigned char cfg_border_width;
    // Keybind *cfg_keybinds;
    // int cfg_kb_count;
    // bool cfg_focus_on_motion;
    Config config;

    int socket_fd;
};

Monitor wm_monitor_open(Display *d, XineramaScreenInfo info);
void wm_monitors_open_all(Wm *wm, Display *d);
Display* wm_connect_display();
void wm_set_supported_atoms(Wm *wm);

void wm_mstack(Wm *wm, Monitor *m);
void wm_set_layout(Wm *wm, void(*f)(Wm *wm, Monitor*));
void wm_layout(Wm *wm, Monitor *m);
bool wm_window_is_dock(Wm* wm, Window w);

NodeType wm_split_to_nodetype(SplitMode mode);
void wm_ws_tree_insert_client(Wm *wm, Workspace *ws, Client *client);

void wm_spawn(Wm* wm, char **str);

void wm_switch_ws(Wm* wm, size_t ws);
void wm_mainloop(Wm* wm);
void wm_init(Wm *wm);
void wm_exit(Wm *wm);

void wm_grab_keys(Wm *wm);
void wm_kb_spawn(Wm *wm, Arg *args);
void wm_kb_kill(Wm *wm, Arg *args);
void wm_kb_switch_ws(Wm *wm, Arg *args);
void wm_kb_focus_dir(Wm *wm, Arg *args);
void wm_kb_move_dir(Wm *wm, Arg *args);
void wm_kb_exit(Wm* wm, Arg *args);
void wm_kb_switch_split_mode(Wm *wm, Arg *args);

struct sockaddr wm_socket_init(Wm *wm);
void wm_socket_cleanup(Wm *wm);
void wm_settings_message_parse(Wm *wm, char *c, int len);

void wm_sigint_handler(Wm *wm, int signum);

#endif
